use std::time::Duration;
use k8s_openapi::api::core::v1::Node;
use kube::{Api, Client};
use kube::api::ListParams;
use proxmox_api::nodes::node::lxc::vmid::migrate;
use proxmox_api::nodes::node::tasks::upid;
use sea_orm::*;
use tokio::time::sleep;
use log::{info, debug};
use proxmox_api::types::VmId;
use crate::api::entity::{locks, nodes};
use crate::api::entity::nodes::{Model as NodeModel, NodeParams, P5xError};
use crate::api::entity::oplogs::Op;
use crate::api::services::Services;


/** WIP: Read nodes from the K8s API and register them w/ P5x based on metadata labels. */
pub async fn discover_nodes(
    svc: &Services<'_>,
) -> Result<(), P5xError> {
    debug!(target: "p5x", "Attempting to discover K8s cluster nodes...");

    let client = Client::try_default().await.map_err(P5xError::KubeError)?;
    let nodes: Api<Node> = Api::all(client);

    let lp = ListParams::default();
    let nodes = nodes.list(&lp).await.map_err(P5xError::KubeError)?;
    for node in nodes {
        let name = node.metadata.name.unwrap_or("(unknown)".to_string());

        // Get the labels for the node
        let labels = node.metadata.labels;
        if labels.is_none() {
            debug!(target: "p5x", "Skipping node {name}: could not load labels");
            continue;
        }
        let labels = labels.unwrap();

        // Check if it has the required pve_host and pve_id labels:
        let pve_host = labels.get("p5x.garrettmills.dev/pve-host");
        let pve_id = labels.get("p5x.garrettmills.dev/pve-id");
        if pve_host.is_none() || pve_id.is_none() {
            debug!(target: "p5x", "Skipping node {name}: missing required pve-host or pve-id label(s)");
            continue;
        }

        let pve_host = pve_host.unwrap();
        let pve_id: Result<i32, _> = pve_id.unwrap().parse();
        if pve_id.is_err() {
            debug!(target: "p5x", "Skipping node {name}: could not parse pve-id into i32");
            continue;
        }
        let pve_id = pve_id.unwrap();

        // Check if the node has been explicitly excluded from p5x management
        let ignore = labels.get("p5x.garrettmills.dev/ignore");
        if ignore.is_some() {
            debug!(target: "p5x", "Skipping node {name}: found explicit ignore label");
            continue;
        }

        // Check if the node is already registered
        let existing = NodeModel::from_pve(svc, pve_host, pve_id).await?;
        if existing.is_some() {
            debug!(target: "p5x", "Skipping node {name}: already registered");
            continue;
        }

        info!(target: "p5x", "Discovered node {name} ({pve_host}:{pve_id}). Attempting to register it.");
        // fixme: WIP
    }

    Ok(())
}


/** Register an existing LXC container with P5x. */
pub async fn register_node(
    svc: &Services<'_>,
    params: &NodeParams,
) -> Result<nodes::Model, P5xError> {
    info!(target: "p5x", "Registering node {} ({})", params.hostname, params.pve_id);

    // Check if the node is already registered
    let existing = nodes::Entity::find()
        .filter(nodes::Column::Hostname.eq(&params.hostname))
        .one(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    if let Some(model) = existing {
        return Ok(model);
    }

    // Insert the new node record
    let model = nodes::ActiveModel {
        hostname: Set(params.hostname.to_string()),
        pve_id: Set(params.pve_id),
        pve_host: Set(params.pve_host.to_string()),
        assigned_ip: Set(params.assigned_ip.to_string()),
        assigned_subnet: Set(params.assigned_subnet),
        is_permanent: Set(params.is_permanent),
        ..Default::default()
    }
        .save(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    // Create the corresponding lock record
    locks::ActiveModel {
        lock_type: Set("nodes".to_string()),
        lock_resource: Set(model.id.clone().unwrap().to_string()),
        ..Default::default()
    }
        .save(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    let node = model.try_into_model()
        .map_err(P5xError::DbErr)?;

    Ok(node)
}


/** Unregister an existing LXC container from P5x. */
pub async fn unregister_node(
    svc: &Services<'_>,
    hostname: &str,
) -> Result<(), P5xError> {
    // Look up the existing node
    let node = nodes::Entity::find()
        .filter(nodes::Column::Hostname.eq(hostname))
        .one(svc.db)
        .await
        .map_err(P5xError::DbErr)?
        .ok_or(P5xError::BadPrecondition("Could not unregister node: unable to find node with that hostname"))?;

    // Acquire the node lock
    let lock = node.lock(svc, Some("Unregistering node from p5x"))
        .await.map_err(P5xError::LockErr)?;

    // Delete the node record first to avoid timing issues w/ locking
    nodes::Entity::delete_by_id(node.id)
        .exec(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    // Release the lock and delete it
    drop(lock);
    locks::Entity::delete_many()
        .filter(locks::Column::LockType.eq("nodes"))
        .filter(locks::Column::LockResource.eq(node.id.to_string()))
        .exec(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    Ok(())
}


/** Migrate an LXC container from its current PVE node to the given PVE node. */
pub async fn migrate_node(
    svc: &Services<'_>,
    node: nodes::Model,
    to_host: &str,
) -> Result<nodes::Model, P5xError> {
    // pre-condition: the node is offline
    if !node.is_offline(svc)? {
        return Err(P5xError::BadPrecondition("Cannot migrate node: node is not offline"))
    }

    // pre-condition: the LXC container has a clean disk config
    if node.config(svc)?.has_pending() {
        return Err(P5xError::BadPrecondition("Cannot migrate node: node has unclean PVE config"))
    }

    svc.op_wrap_f(
        svc.op_start(Op::MigrateNode, &format!("Migrate node {} ({} to {})", node.vm_id(), &node.pve_host, to_host)).await?,
        "Migrated node",
        async || {
            migrate_node_unmanaged(svc, &node.pve_host, node.pve_id, to_host).await
        }
    ).await?;

    // Persist the node
    let mut node = node.into_active_model();
    node.pve_host = Set(to_host.to_string());
    let node = node.save(svc.db).await.map_err(P5xError::DbErr)?;

    nodes::Entity::find_by_id(node.id.unwrap())
        .one(svc.db)
        .await
        .map_err(P5xError::DbErr)?
        .ok_or(P5xError::BadPostcondition("Could not look up node after persisting"))
}


/** Like migrate_node, but without any of the SQL bookkeeping, locking, or safety checks. */
pub async fn migrate_node_unmanaged(
    svc: &Services<'_>,
    from_pve_host: &str,
    from_pve_id: i32,
    to_host: &str,
) -> Result<(), P5xError> {
    // Ask the PVE API to start migrating the node
    let params = migrate::PostParams::new(to_host.to_string());
    let upid = svc.pve_node(from_pve_host)
        .map_err(P5xError::ServiceError)?
        .lxc()
        .vmid(VmId::new(i64::from(from_pve_id)).unwrap())
        .migrate()
        .post(params)
        .map_err(P5xError::PveError)?;

    // post-condition: the LXC container migrated successfully
    wait_upid(svc, from_pve_host, &upid).await
}


/** Wait for a PVE task to complete using its UPID */
pub async fn wait_upid(svc: &Services<'_>, node: &str, upid: &str) -> Result<(), P5xError> {
    info!("Waiting for UPID {upid} on node {node}");
    let pve = svc.pve_node(node)
        .map_err(P5xError::ServiceError)?;

    loop {
        let status = pve.tasks()
            .upid(upid)
            .status()
            .get()
            .map_err(P5xError::PveError)?;

        if status.status == upid::status::Status::Running {
            sleep(Duration::from_secs(1)).await;
            continue;
        }

        if let Some(s) = status.exitstatus {
            if s == "OK" {
                info!("UPID {upid} on node {node} finished");
                return Ok(());
            }

            error!("UPID {upid} on node {node} failed");
            return Err(P5xError::UpidFailed(node.to_string(), upid.to_string()));
        }
    }
}
